from tempfile import TemporaryDirectory
from time import sleep


class AltTemporaryDirectory:
    def __init__(self):
        self.tmpdir = TemporaryDirectory()

    def __enter__(self):
        return self.tmpdir.name

    def cleanup(self, cnt=0):
        if cnt >= 5:  # pragma: no cover
            raise RuntimeError("Could not delete TemporaryDirectory!")
        try:
            self.tmpdir.cleanup()
        except IOError:  # pragma: no cov_4_nix
            sleep(1)
            self.cleanup(cnt + 1)

    def __exit__(self, exc, value, tb):
        self.cleanup()



import json
from pathlib import Path
from subprocess import DEVNULL, PIPE, CalledProcessError, run  # nosec
from typing import Optional, Set

import click
import typer

from . import utils
from .utils import example
from .utils.iohelper import AltTemporaryDirectory


@example(skip_apply_ask=False)
@example()
def update(
    project_dir: Path = Path("."),
    cookiecutter_input: bool = False,
    refresh_private_variables: bool = False,
    skip_apply_ask: bool = True,
    skip_update: bool = False,
    checkout: Optional[str] = None,
    strict: bool = True,
    allow_untracked_files: bool = False,
) -> bool:
    """Update specified project's cruft to the latest and greatest release."""
    cruft_file = utils.cruft.get_cruft_file(project_dir)

    # If the project dir is a git repository, we ensure
    # that the user has a clean working directory before proceeding.
    if not _is_project_repo_clean(project_dir, allow_untracked_files):
        typer.secho(
            "Cruft cannot apply updates on an unclean git project."
            " Please make sure your git working tree is clean before proceeding.",
            fg=typer.colors.RED,
        )
        return False

    cruft_state = json.loads(cruft_file.read_text())

    with AltTemporaryDirectory() as tmpdir_:
        # Initial setup
        tmpdir = Path(tmpdir_)
        repo_dir = tmpdir / "repo"
        current_template_dir = tmpdir / "current_template"
        new_template_dir = tmpdir / "new_template"
        deleted_paths: Set[Path] = set()
        # Clone the template
        with utils.cookiecutter.get_cookiecutter_repo(
            cruft_state["template"], repo_dir, checkout
        ) as repo:
            last_commit = repo.head.object.hexsha

            # Bail early if the repo is already up to date and no inputs are asked
            if not (
                cookiecutter_input or refresh_private_variables
            ) and utils.cruft.is_project_updated(repo, cruft_state["commit"], last_commit, strict):
                typer.secho(
                    "Nothing to do, project's cruft is already up to date!", fg=typer.colors.GREEN
                )
                return True

            # Generate clean outputs via the cookiecutter
            # from the current cruft state commit of the cookiectter and the updated
            # cookiecutter.
            # For the current cruft state, we do not try to update the cookiecutter_input
            # because we want to keep the current context input intact.
            _ = utils.generate.cookiecutter_template(
                output_dir=current_template_dir,
                repo=repo,
                cruft_state=cruft_state,
                project_dir=project_dir,
                checkout=cruft_state["commit"],
                deleted_paths=deleted_paths,
                update_deleted_paths=True,
            )
            # Remove private variables from cruft_state to refresh their values
            # from the cookiecutter template config
            if refresh_private_variables:
                _clean_cookiecutter_private_variables(cruft_state)

            new_context = utils.generate.cookiecutter_template(
                output_dir=new_template_dir,
                repo=repo,
                cruft_state=cruft_state,
                project_dir=project_dir,
                cookiecutter_input=cookiecutter_input,
                checkout=last_commit,
                deleted_paths=deleted_paths,
            )

        # Given the two versions of the cookiecutter outputs based
        # on the current project's context we calculate the diff and
        # apply the updates to the current project.
        if _apply_project_updates(
            current_template_dir,
            new_template_dir,
            project_dir,
            skip_update,
            skip_apply_ask,
            allow_untracked_files,
        ):
            # Update the cruft state and dump the new state
            # to the cruft file
            cruft_state["commit"] = last_commit
            cruft_state["checkout"] = checkout
            cruft_state["context"] = new_context
            cruft_file.write_text(utils.cruft.json_dumps(cruft_state))
            typer.secho(
                "Good work! Project's cruft has been updated and is as clean as possible!",
                fg=typer.colors.GREEN,
            )
        return True


def _clean_cookiecutter_private_variables(cruft_state: dict):
    for key in list(cruft_state["context"]["cookiecutter"].keys()):
        if key != "_template" and key.startswith("_"):
            del cruft_state["context"]["cookiecutter"][key]


#################################################
# Calculating project diff and applying updates #
#################################################


def _is_git_repo(directory: Path):
    # Taken from https://stackoverflow.com/a/16925062
    # This works even if we are in a sub folder in a git
    # repo
    output = run(
        ["git", "rev-parse", "--is-inside-work-tree"], stdout=PIPE, stderr=DEVNULL, cwd=directory
    )
    if b"true" in output.stdout:
        return True
    return False


def _has_untracked_file(status_line: str):
    return status_line.strip().startswith("??")


def _is_project_repo_clean(directory: Path, allow_untracked_files: bool):
    if not _is_git_repo(directory):
        return True
    git_status = run(["git", "status", "--porcelain"], stdout=PIPE, stderr=DEVNULL, cwd=directory)
    status_lines = git_status.stdout.decode("utf-8").split("\n")
    # remove empty string from trailing newline
    status_lines = [line for line in status_lines if line]
    if allow_untracked_files:
        status_lines = [line for line in status_lines if not _has_untracked_file(line)]
    if status_lines:
        return False
    return True


def _apply_patch_with_rejections(diff: str, expanded_dir_path: Path):
    offset = _get_offset(expanded_dir_path)

    git_apply = ["git", "apply", "--reject"]
    if offset:
        git_apply.extend(["--directory", offset])

    try:
        run(
            git_apply,
            input=diff.encode(),
            stderr=PIPE,
            stdout=PIPE,
            check=True,
            cwd=expanded_dir_path,
        )
    except CalledProcessError as error:
        typer.secho(error.stderr.decode(), err=True)
        typer.secho(
            (
                "Project directory may have *.rej files reflecting merge conflicts with the update."
                " Please resolve those conflicts manually."
            ),
            fg=typer.colors.YELLOW,
        )


def _apply_three_way_patch(diff: str, expanded_dir_path: Path, allow_untracked_files: bool):
    offset = _get_offset(expanded_dir_path)

    git_apply = ["git", "apply", "-3"]
    if offset:
        git_apply.extend(["--directory", offset])

    try:
        run(
            git_apply,
            input=diff.encode(),
            stderr=PIPE,
            stdout=PIPE,
            check=True,
            cwd=expanded_dir_path,
        )
    except CalledProcessError as error:
        typer.secho(error.stderr.decode(), err=True)
        if _is_project_repo_clean(expanded_dir_path, allow_untracked_files):
            typer.secho(
                "Failed to apply the update. Retrying again with a different update strategy.",
                fg=typer.colors.YELLOW,
            )
            _apply_patch_with_rejections(diff, expanded_dir_path)


def _get_offset(expanded_dir_path: Path):
    try:
        offset = (
            run(
                ["git", "rev-parse", "--show-prefix"],
                stderr=PIPE,
                stdout=PIPE,
                check=True,
                cwd=expanded_dir_path,
            )
            .stdout.decode()
            .strip()
        )
        return offset
    except CalledProcessError as error:
        if "not a git repository" in error.stderr.decode():
            return ""
        else:
            raise error


def _apply_patch(diff: str, expanded_dir_path: Path, allow_untracked_files: bool):
    # Git 3 way merge is the our best bet
    # at applying patches. But it only works
    # with git repos. If the repo is not a git dir
    # we fall back to git apply --reject which applies
    # diffs cleanly where applicable otherwise creates
    # *.rej files where there are conflicts
    if _is_git_repo(expanded_dir_path):
        _apply_three_way_patch(diff, expanded_dir_path, allow_untracked_files)
    else:
        _apply_patch_with_rejections(diff, expanded_dir_path)


def _apply_project_updates(
    old_main_directory: Path,
    new_main_directory: Path,
    project_dir: Path,
    skip_update: bool,
    skip_apply_ask: bool,
    allow_untracked_files: bool,
) -> bool:
    diff = utils.diff.get_diff(old_main_directory, new_main_directory)

    if not skip_apply_ask and not skip_update:
        input_str: str = "v"
        while input_str == "v":
            typer.echo(
                'Respond with "s" to intentionally skip the update while marking '
                "your project as up-to-date or "
                'respond with "v" to view the changes that will be applied.'
            )
            input_str = typer.prompt(
                "Apply diff and update?",
                type=click.Choice(("y", "n", "s", "v")),
                show_choices=True,
                default="y",
            )
            if input_str == "v":
                if diff.strip():
                    utils.diff.display_diff(old_main_directory, new_main_directory)
                else:
                    click.secho("There are no changes.", fg=typer.colors.YELLOW)
        if input_str == "n":
            typer.echo("User cancelled Cookiecutter template update.")
            return False
        elif input_str == "s":
            skip_update = True

    if not skip_update and diff.strip():
        _apply_patch(diff, project_dir, allow_untracked_files)
    return True



"""Contains all custom exceptions raised by cruft."""
from pathlib import Path
from typing import Union

from click import ClickException


class CruftError(ClickException):
    """The base exception for any error originating from the cruft project."""


class UnableToFindCookiecutterTemplate(CruftError):
    """Raised when Cruft is unable to find a cookiecutter template."""

    def __init__(self, directory: Union[str, Path]):
        if not isinstance(directory, str):
            directory = str(directory)
        super().__init__(f"Was unable to locate a Cookiecutter template in `{directory}` !")
        self.directory = directory


class NoCruftFound(CruftError):
    """Raised when no .cruft.json state is found in the current directory."""

    def __init__(self, directory: Union[str, Path]):
        if not isinstance(directory, str):
            directory = str(directory)
        super().__init__(f"Was unable to locate a `.cruft.json` state file in `{directory}` !")
        self.directory = directory


class CruftAlreadyPresent(CruftError):
    """Raised when there is an attempt to create a new .cruft.json file but one already exists."""

    def __init__(self, file_location: Union[str, Path]):
        if not isinstance(file_location, str):
            file_location = str(file_location)
        super().__init__(f"`.cruft.json` is already defined at `{file_location}` !")
        self.file_location = file_location


class InvalidCookiecutterRepository(CruftError):
    """Raised when an invalid cookiecutter repository is provided."""

    def __init__(self, cookiecutter_repo: str, details: str = ""):
        self.cookiecutter_repo = cookiecutter_repo
        super().__init__(
            f"Unable to initialize the cookiecutter using {cookiecutter_repo}! {details.strip()}"
        )


class ChangesetUnicodeError(CruftError):
    """Raised when `cruft update` is unable to generate the change"""

    def __init__(self):
        super().__init__(
            (
                "Unable to interpret changes between current project and cookiecutter template as "
                "unicode. Typically a result of hidden binary files in project folder."
            )
        )



"""Contains the core logic behind all cruft commands."""
from .check import check
from .create import create
from .diff import diff
from .link import link
from .update import update

__all__ = ["check", "create", "diff", "link", "update"]



from pathlib import Path
from typing import Any, Dict, Optional
from urllib.parse import urlparse

from cookiecutter.config import get_user_config
from cookiecutter.generate import generate_context
from cookiecutter.prompt import prompt_for_config
from git import GitCommandError, Repo

from cruft.exceptions import InvalidCookiecutterRepository, UnableToFindCookiecutterTemplate

CookiecutterContext = Dict[str, Any]


#################################
# Cookiecutter helper functions #
#################################


def resolve_template_url(url: str) -> str:
    parsed_url = urlparse(url)
    # If we are given a file URI, we should convert
    # relative paths to absolute paths. This is to
    # make sure that further operations like check/update
    # work properly in case the generated project directory
    # does not reside in the same relative path.
    if not parsed_url.scheme or parsed_url.scheme == "file":
        file_path = (Path(parsed_url.netloc) / Path(parsed_url.path)).absolute()
        # Below is to handle cases like "git@github.com"
        # which passes through to this block, but will obviously not
        # exist in the file system.
        # In this case we simply return the URL. If the user did
        # pass in a valid file path that does not exist, we do not need to
        # worry as we will never to be able use it in check/update etc. anyway
        if file_path.exists():
            return str(file_path)
    return url


def get_cookiecutter_repo(
    template_git_url: str,
    cookiecutter_template_dir: Path,
    checkout: Optional[str] = None,
    **clone_kwargs,
) -> Repo:
    try:
        repo = Repo.clone_from(template_git_url, cookiecutter_template_dir, **clone_kwargs)
    except GitCommandError as error:
        raise InvalidCookiecutterRepository(
            template_git_url, f"Failed to clone the repo. {error.stderr.strip()}"
        )
    if checkout is not None:
        try:
            repo.git.checkout(checkout)
        except GitCommandError as error:
            raise InvalidCookiecutterRepository(
                template_git_url,
                f"Failed to check out the reference {checkout}. {error.stderr.strip()}",
            )
    return repo


def _validate_cookiecutter(cookiecutter_template_dir: Path):
    main_cookiecutter_directory: Optional[Path] = None

    for dir_item in cookiecutter_template_dir.glob("*cookiecutter.*"):
        if dir_item.is_dir() and "{{" in dir_item.name and "}}" in dir_item.name:
            main_cookiecutter_directory = dir_item
            break

    if not main_cookiecutter_directory:
        raise UnableToFindCookiecutterTemplate(cookiecutter_template_dir)


def generate_cookiecutter_context(
    template_git_url: str,
    cookiecutter_template_dir: Path,
    config_file: Optional[Path] = None,
    default_config: bool = False,
    extra_context: Optional[Dict[str, Any]] = None,
    no_input: bool = False,
) -> CookiecutterContext:
    _validate_cookiecutter(cookiecutter_template_dir)

    context_file = cookiecutter_template_dir / "cookiecutter.json"
    config_dict = get_user_config(
        config_file=str(config_file) if config_file else None, default_config=default_config
    )

    context = generate_context(
        context_file=context_file,
        default_context=config_dict["default_context"],
        extra_context=extra_context,
    )

    # prompt the user to manually configure at the command line.
    # except when 'no-input' flag is set
    context["cookiecutter"] = prompt_for_config(context, no_input)
    context["cookiecutter"]["_template"] = template_git_url

    return context



from pathlib import Path
from typing import Any, Dict, List, Optional

from cookiecutter.generate import generate_files

from . import utils
from .utils import example
from .utils.iohelper import AltTemporaryDirectory


@example("https://github.com/timothycrosley/cookiecutter-python/")
def create(
    template_git_url: str,
    output_dir: Path = Path("."),
    config_file: Optional[Path] = None,
    default_config: bool = False,
    extra_context: Optional[Dict[str, Any]] = None,
    no_input: bool = True,
    directory: Optional[str] = None,
    checkout: Optional[str] = None,
    overwrite_if_exists: bool = False,
    skip: Optional[List[str]] = None,
) -> Path:
    """Expand a Git based Cookiecutter template into a new project on disk."""
    template_git_url = utils.cookiecutter.resolve_template_url(template_git_url)
    with AltTemporaryDirectory() as cookiecutter_template_dir_str:
        cookiecutter_template_dir = Path(cookiecutter_template_dir_str)
        with utils.cookiecutter.get_cookiecutter_repo(
            template_git_url, cookiecutter_template_dir, checkout
        ) as repo:
            last_commit = repo.head.object.hexsha

            if directory:
                cookiecutter_template_dir = cookiecutter_template_dir / directory

            context = utils.cookiecutter.generate_cookiecutter_context(
                template_git_url,
                cookiecutter_template_dir,
                config_file,
                default_config,
                extra_context,
                no_input,
            )

        project_dir = Path(
            generate_files(
                repo_dir=cookiecutter_template_dir,
                context=context,
                overwrite_if_exists=overwrite_if_exists,
                output_dir=str(output_dir),
            )
        )

        cruft_content = {
            "template": template_git_url,
            "commit": last_commit,
            "checkout": checkout,
            "context": context,
            "directory": directory,
        }

        if skip:
            cruft_content["skip"] = skip

        # After generating the project - save the cruft state
        # into the cruft file.
        (project_dir / ".cruft.json").write_text(utils.cruft.json_dumps(cruft_content))

        return project_dir



from pathlib import Path
from re import sub
from subprocess import PIPE, run  # nosec
from typing import List

from cruft import exceptions

DIFF_SRC_PREFIX = "upstream-template-old"
DIFF_DST_PREFIX = "upstream-template-new"


def _git_diff(*args: str) -> List[str]:
    # https://git-scm.com/docs/git-diff#Documentation/git-diff.txt---binary support for binary patch
    return [
        "git",
        "-c",
        "diff.noprefix=",
        "diff",
        "--no-index",
        "--relative",
        "--binary",
        f"--src-prefix={DIFF_SRC_PREFIX}/",
        f"--dst-prefix={DIFF_DST_PREFIX}/",
        *args,
    ]


def get_diff(repo0: Path, repo1: Path) -> str:
    """Compute the raw diff between two repositories."""
    # Use Path methods in order to straighten out the differences between the the OSs.
    repo0_str = repo0.resolve().as_posix()
    repo1_str = repo1.resolve().as_posix()
    try:
        diff = run(
            _git_diff("--no-ext-diff", "--no-color", repo0_str, repo1_str),
            cwd=repo0_str,
            stdout=PIPE,
            stderr=PIPE,
        ).stdout.decode()
    except UnicodeDecodeError:
        raise exceptions.ChangesetUnicodeError()
    # By default, git diff --no-index will output full paths like so:
    # --- a/tmp/tmpmp34g21y/remote/.coveragerc
    # +++ b/tmp/tmpmp34g21y/local/.coveragerc
    # We don't want this as we may need to apply the diff later on.
    # Note that diff headers contain repo0 and repo1 with both "a" and "b"
    # prefixes: headers for new files have a/repo1, headers for deleted files
    # have b/repo0.
    # NIX OPs have a/folder/file
    # WIN OPS have a/c:/folder/file
    # More info on git-diff can be found here: http://git-scm.com/docs/git-diff
    for repo in [repo0_str, repo1_str]:
        # Make repo look like a NIX absolute path.
        repo = sub("/[a-z]:", "", repo)
        diff = diff.replace(f"{DIFF_SRC_PREFIX}{repo}", DIFF_SRC_PREFIX).replace(
            f"{DIFF_DST_PREFIX}{repo}", DIFF_DST_PREFIX
        )

    # This replacement is needed for renamed/moved files to be recognized properly
    # Renamed files in the diff don't have the "a" or "b" prefix and instead look like
    # /tmp/tmpmp34g21y/remote/.coveragerc
    # If we replace repo paths which are like /tmp/tmpmp34g21y/remote
    # we would end up with /.coveragerc which doesn't work.
    # We also need to replace the trailing slash. As a result, we only do
    # this after the above replacement is made as the trailing slash is needed there.
    diff = diff.replace(repo0_str + "/", "").replace(repo1_str + "/", "")

    return diff


def display_diff(repo0: Path, repo1: Path):
    """Displays the diff between two repositories."""
    run(_git_diff(repo0.as_posix(), repo1.as_posix()))



from functools import wraps

from . import cookiecutter, cruft, diff, generate, iohelper

try:
    from examples import example
except ImportError:  # pragma: no cover
    # In case examples is not available,
    # we introduce a no-op decorator.
    def example(*_args, **_kwargs):
        def decorator(f):
            @wraps(f)
            def wrapper(*args, **kwargs):
                return f(*args, **kwargs)

            return wrapper

        return decorator


__all__ = ["cookiecutter", "cruft", "diff", "example", "generate", "iohelper"]



import json
import shutil
import sys
from pathlib import Path
from typing import Optional

import typer

from . import utils
from .utils.iohelper import AltTemporaryDirectory


def diff(
    project_dir: Path = Path("."), exit_code: bool = False, checkout: Optional[str] = None
) -> bool:
    """Show the diff between the project and the linked Cookiecutter template"""
    cruft_file = utils.cruft.get_cruft_file(project_dir)
    cruft_state = json.loads(cruft_file.read_text())
    checkout = checkout or cruft_state.get("commit")

    has_diff = False
    with AltTemporaryDirectory() as tmpdir_:
        tmpdir = Path(tmpdir_)
        repo_dir = tmpdir / "repo"
        remote_template_dir = tmpdir / "remote"
        local_template_dir = tmpdir / "local"

        # Create all the directories
        remote_template_dir.mkdir(parents=True, exist_ok=True)
        local_template_dir.mkdir(parents=True, exist_ok=True)

        # Let's clone the template
        with utils.cookiecutter.get_cookiecutter_repo(
            cruft_state["template"], repo_dir, checkout=checkout
        ) as repo:

            # We generate the template for the revision expected by the project
            utils.generate.cookiecutter_template(
                output_dir=remote_template_dir,
                repo=repo,
                cruft_state=cruft_state,
                project_dir=project_dir,
                checkout=checkout,
                update_deleted_paths=True,
            )

        # Then we create a new tree with each file in the template that also exist
        # locally.
        for path in sorted(remote_template_dir.glob("**/*")):
            relative_path = path.relative_to(remote_template_dir)
            local_path = project_dir / relative_path
            destination = local_template_dir / relative_path
            if path.is_file():
                shutil.copy(str(local_path), str(destination))
            else:
                destination.mkdir(parents=True, exist_ok=True)
                destination.chmod(local_path.stat().st_mode)

        # Finally we can compute and print the diff.
        diff = utils.diff.get_diff(local_template_dir, remote_template_dir)

        if diff.strip():
            has_diff = True

            if exit_code or not sys.stdout.isatty():
                # The current shell doesn't run on a TTY or the "--exit-code" flag
                # is set. This means we're probably not displaying the diff to an
                # end-user. Let's just output the sanitized version of the diff.
                #
                # Note that we can't delegate this check to "git diff" command
                # because it would show absolute paths to files as we're working in
                # temporary, non-gitted directories. Doing so would prevent the user
                # from applying the patch later on as the temporary directories wouldn't
                # exist anymore.
                typer.echo(diff, nl=False)
            else:
                # We're outputing the diff to a real user. We can delegate the job
                # to git diff so that they can benefit from coloration and paging.
                # Ouputing absolute paths is less of a concern although it would be
                # better to find a way to make git shrink those paths.
                utils.diff.display_diff(local_template_dir, remote_template_dir)

    return not (has_diff and exit_code)



import json
from pathlib import Path
from typing import Any, Dict

from git import Repo

from cruft.exceptions import CruftAlreadyPresent, NoCruftFound

CruftState = Dict[str, Any]


#######################
# Cruft related utils #
#######################


def get_cruft_file(project_dir_path: Path, exists: bool = True) -> Path:
    cruft_file = project_dir_path / ".cruft.json"
    if not exists and cruft_file.is_file():
        raise CruftAlreadyPresent(cruft_file)
    if exists and not cruft_file.is_file():
        raise NoCruftFound(project_dir_path.resolve())
    return cruft_file


def is_project_updated(repo: Repo, current_commit: str, latest_commit: str, strict: bool) -> bool:
    return (
        # If the latest commit exactly matches the current commit
        latest_commit == current_commit
        # Or if there have been no changes to the cookiecutter
        or not repo.index.diff(current_commit)
        # or if the strict flag is off, we allow for newer commits to count as up to date
        or (
            repo.is_ancestor(repo.commit(latest_commit), repo.commit(current_commit)) and not strict
        )
    )


def json_dumps(cruft_state: Dict[str, Any]) -> str:
    text = json.dumps(cruft_state, ensure_ascii=False, indent=2, separators=(",", ": "))
    return text + "\n"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cruft', 'cruft._commands', 'cruft._commands.utils']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.2', 'cookiecutter>=1.7', 'gitpython>=3.0', 'typer>=0.4.0']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=2.0.0,<5.0.0'],
 'pyproject': ['toml>=0.10']}

entry_points = \
{'console_scripts': ['cruft = cruft._cli:app']}

setup_kwargs = {
    'name': 'cruft',
    'version': '2.12.0',
    'description': 'Allows you to maintain all the necessary cruft for packaging and building projects separate from the code you intentionally write. Built on-top of CookieCutter.',
    'long_description': '[![cruft - Fight Back Against the Boilerplate Monster!](https://raw.github.com/cruft/cruft/master/art/logo_large.png)](https://cruft.github.io/cruft/)\n_________________\n\n[![PyPI version](https://badge.fury.io/py/cruft.svg)](http://badge.fury.io/py/cruft)\n[![Conda Version](https://img.shields.io/conda/vn/conda-forge/cruft.svg)](https://anaconda.org/conda-forge/cruft)\n[![Build Status](https://github.com/cruft/cruft/workflows/Run%20tests/badge.svg)](https://github.com/cruft/cruft/actions?query=workflow%3A%22Run+tests%22+branch%3Amaster)\n[![codecov](https://codecov.io/gh/cruft/cruft/branch/master/graph/badge.svg)](https://codecov.io/gh/cruft/cruft)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Imports: isort](https://img.shields.io/badge/%20imports-isort-%231674b1?style=flat&labelColor=ef8336)](https://timothycrosley.github.io/isort/)\n[![Join the chat at https://gitter.im/cruft/community](https://badges.gitter.im/cruft/community.svg)](https://gitter.im/cruft/community?utm_source=badge&utm_medium=badge&utm_campaign=pr-badge&utm_content=badge)\n[![License](https://img.shields.io/github/license/mashape/apistatus.svg)](https://pypi.python.org/pypi/cruft/)\n[![Downloads](https://pepy.tech/badge/cruft)](https://pepy.tech/project/cruft)\n\n#### Trending Contributors\n\n[![](https://sourcerer.io/fame/samj1912/cruft/cruft/images/0)](https://sourcerer.io/fame/samj1912/cruft/cruft/links/0)[![](https://sourcerer.io/fame/samj1912/cruft/cruft/images/1)](https://sourcerer.io/fame/samj1912/cruft/cruft/links/1)[![](https://sourcerer.io/fame/samj1912/cruft/cruft/images/2)](https://sourcerer.io/fame/samj1912/cruft/cruft/links/2)[![](https://sourcerer.io/fame/samj1912/cruft/cruft/images/3)](https://sourcerer.io/fame/samj1912/cruft/cruft/links/3)[![](https://sourcerer.io/fame/samj1912/cruft/cruft/images/4)](https://sourcerer.io/fame/samj1912/cruft/cruft/links/4)[![](https://sourcerer.io/fame/samj1912/cruft/cruft/images/5)](https://sourcerer.io/fame/samj1912/cruft/cruft/links/5)[![](https://sourcerer.io/fame/samj1912/cruft/cruft/images/6)](https://sourcerer.io/fame/samj1912/cruft/cruft/links/6)[![](https://sourcerer.io/fame/samj1912/cruft/cruft/images/7)](https://sourcerer.io/fame/samj1912/cruft/cruft/links/7)\n_________________\n\n[Read Latest Documentation](https://cruft.github.io/cruft/) - [Browse GitHub Code Repository](https://github.com/cruft/cruft/)\n_________________\n\n**cruft** allows you to maintain all the necessary boilerplate for packaging and building projects separate from the code you intentionally write.\nFully compatible with existing [Cookiecutter](https://github.com/cookiecutter/cookiecutter) templates.\n\nCreating new projects from templates using cruft is easy:\n\n![Example Usage New Project](https://raw.githubusercontent.com/cruft/cruft/master/art/example.gif)\n\nAnd, so is updating them as the template changes overtime:\n\n![Example Usage New Project](https://raw.githubusercontent.com/cruft/cruft/master/art/example_update.gif)\n\nMany project template utilities exist that automate the copying and pasting of code to create new projects. This *seems* great! However, once created, most leave you with that copy-and-pasted code to manage through the life of your project.\n\ncruft is different. It automates the creation of new projects like the others, but then it also helps you to manage the boilerplate through the life of the project. cruft makes sure your code stays in-sync with the template it came from for you.\n\n## Key Features:\n\n* **Cookiecutter Compatible**: cruft utilizes [Cookiecutter](https://github.com/cookiecutter/cookiecutter) as its template expansion engine. Meaning it retains full compatibility with all existing [Cookiecutter](https://github.com/cookiecutter/cookiecutter) templates.\n* **Template Validation**: cruft can quickly validate whether or not a project is using the latest version of a template using `cruft check`. This check can easily be added to CI pipelines to ensure your projects stay in-sync.\n* **Automatic Template Updates**: cruft automates the process of updating code to match the latest version of a template, making it easy to utilize template improvements across many projects.\n\n## Installation:\n\nTo get started - install `cruft` using a Python package manager:\n\n`pip3 install cruft`\n\nOR\n\n`poetry add cruft`\n\nOR\n\n`pipenv install cruft`\n\n\n## Creating a New Project:\n\nTo create a new project using cruft run `cruft create PROJECT_URL` from the command line.\n\nFor example:\n\n        cruft create https://github.com/timothycrosley/cookiecutter-python/\n\ncruft will then ask you any necessary questions to create your new project. It will use your answers to expand the provided template, and then return the directory it placed the expanded project.\nBehind the scenes, cruft uses [Cookiecutter](https://github.com/cookiecutter/cookiecutter) to do the project expansion. The only difference in the resulting output is a `.cruft.json` file that\ncontains the git hash of the template used as well as the parameters specified.\n\n## Updating a Project\n\nTo update an existing project, that was created using cruft, run `cruft update` in the root of the project.\nIf there are any updates, cruft will have you review them before applying. If you accept the changes cruft will apply them to your project\nand update the `.cruft.json` file for you.\n\n!!! tip\n    Sometimes certain files just aren\'t good fits for updating. Such as test cases or `__init__` files. You can tell cruft to always skip updating these files on a project by generating project with `--skip cruft/__init__.py --skip tests` arguments or manually adding them to a skip section within your `.cruft.json` file:\n\n        {\n            "template": "https://github.com/timothycrosley/cookiecutter-python",\n            "commit": "8a65a360d51250221193ed0ec5ed292e72b32b0b",\n            "skip": [\n                "cruft/__init__.py",\n                "tests"\n            ],\n            ...\n        }\n\n    Or, if you have toml installed, you can add skip files directly to a `tool.cruft` section of your `pyproject.toml` file:\n\n        [tool.cruft]\n        skip = ["cruft/__init__.py", "tests"]\n    \n    Note that it is possible to use glob patterns for selecting the files to skip:\n        {\n            "skip": [\n                "**/__init__.py",\n                "tests/*"\n            ],\n            ...\n        }\n\n\n\n## Checking a Project\n\nChecking to see if a project is missing a template update is as easy as running `cruft check`. If the project is out-of-date an error and exit code 1 will be returned.\n`cruft check` can be added to CI pipelines to ensure projects don\'t unintentionally drift.\n\n\n## Linking an Existing Project\n\nHave an existing project that you created from a template in the past using Cookiecutter directly? You can link it to the template that was used to create it using: `cruft link TEMPLATE_REPOSITORY`.\n\nFor example:\n\n        cruft link https://github.com/timothycrosley/cookiecutter-python/\n\nYou can then specify the last commit of the template the project has been updated to be consistent with, or accept the default of using the latest commit from the template.\n\n## Compute the diff\n\nWith time, your boilerplate may end up being very different from the actual cookiecutter template. Cruft allows you to quickly see what changed in your local project compared to the template. It is as easy as running `cruft diff`. If any local file differs from the template, the diff will appear in your terminal in a similar fashion to `git diff`.\n\nThe `cruft diff` command optionally accepts an `--exit-code` flag that will make cruft exit with a non-0 code should any diff is found. You can combine this flag with the `skip` section of your `.cruft.json` to make stricter CI checks that ensures any improvement to the template is always submitted upstream.\n\n## Why Create cruft?\n\nSince I first saw videos of [quickly](https://www.youtube.com/watch?v=9EctXzH2dss) being used to automate Ubuntu application creation, I\'ve had a love/hate relationship with these kinds of tools.\nI\'ve used them for many projects and certainly seen them lead to productivity improvements. However, I\'ve always felt like they were a double-edged sword. Sure, they would automate away the copying and pasting many would do to create projects. However, by doing so,\nthey encouraged more code to be copied and pasted! Then, over time, you could easily be left with hundreds of projects that contained copy-and-pasted code with no way to easy way to update them. I created cruft to be a tool that recognized that balance between project creation and maintenance and provided mechanisms to keep built projects up-to-date.\n\nI hope you too find `cruft` useful!\n\n~Timothy Crosley\n',
    'author': 'Timothy Crosley',
    'author_email': 'timothy.crosley@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)



import json
from pathlib import Path
from typing import Optional

import typer

from . import utils
from .utils import example
from .utils.iohelper import AltTemporaryDirectory


@example()
def check(
    project_dir: Path = Path("."), checkout: Optional[str] = None, strict: bool = True
) -> bool:
    """Checks to see if there have been any updates to the Cookiecutter template
    used to generate this project."""
    cruft_file = utils.cruft.get_cruft_file(project_dir)
    cruft_state = json.loads(cruft_file.read_text())
    with AltTemporaryDirectory() as cookiecutter_template_dir:
        with utils.cookiecutter.get_cookiecutter_repo(
            cruft_state["template"],
            Path(cookiecutter_template_dir),
            checkout,
            filter="blob:none",
            no_checkout=True,
        ) as repo:
            last_commit = repo.head.object.hexsha

            if utils.cruft.is_project_updated(repo, cruft_state["commit"], last_commit, strict):
                typer.secho(
                    "SUCCESS: Good work! Project's cruft is up to date "
                    "and as clean as possible :).",
                    fg=typer.colors.GREEN,
                )
                return True

        typer.secho(
            "FAILURE: Project's cruft is out of date! Run `cruft update` to clean this mess up.",
            fg=typer.colors.RED,
        )
        return False



"""**cruft**

Allows you to maintain all the necessary cruft for packaging and building projects separate from
the code you intentionally write. Built on-top of, and fully compatible with, CookieCutter.
"""
from cruft._commands import check, create, diff, link, update
from cruft._version import __version__

__all__ = ["create", "check", "diff", "update", "link", "__version__"]
